import boto3
import json
import uuid

from api import Response
from dynamodb import Table

def handler(event, context):
    """Runs a image"""

    method = event['httpMethod']
    resource = event['resource']

    key = (event.get('pathParameters') or {}).get('id', None)
    body = event.get('body', {})

    functions = {
        'GET': {
            '/execution/{id}': detail,
            '/execution': list,
        },
        'POST': {
            '/execution': create,
        },
        'DELETE': {
            '/execution/{id}': delete,
        },
        'PATCH': {
            '/execution/{id}': update,
        },
    }

    handler = functions.get(method, {}).get(resource)

    return handler(key=key, body=body)


def detail(key, *args, **kwargs):
    """Describes a entry"""

    table = Table('REGISTRY_TABLE_NAME')

    response = Response(200, table.get(key))

    return response.parse()

def list(*args, **kwargs):
    """Lists all entries"""
    pass

def delete(key, *args, **kwargs):
    """Deletes a entry"""
    pass

def update(key, body, *args, **kwargs):
    """Updates (patches) a entry"""
    pass

def create(body, *args, **kwargs):
    """Creates a entry"""

    table = Table('JOBS_TABLE_NAME')
    registry = Table('REGISTRY_TABLE_NAME')
    client = boto3.client('ecs')

    # Parses the body
    body = json.loads(body)

    # Generates a uuid and updates the table
    id = uuid.uuid4().hex
    table.update(id, {'status': 'RECEIVED'})

    # Gets the task definition from the registry
    image = registry.get(body['image'])

    # Image not built yet
    if image is None:

        return Response(404, 'Image does not exists.').parse()

    elif image['status'] != 'SUCCEEDED':

        return Response(409, 'Image is not built yet.').parse()

    # Default parameters
    parameters = {
        'cpu': body.get('cpu', image['definition']['cpu']),
        'memory': body.get('memory', image['definition']['memory']),
        'count': body.get('count', 1),
        'role': body.get('role', image['definition']['roles']['task']),
        'environment': body.get('environment', []),
    }

    response = client.run_task(
        cluster='harbour',
        count=parameters['count'],
        launchType='FARGATE',
        group=id,
        networkConfiguration={
            'awsvpcConfiguration': {
                'subnets': [
                    'subnet-00f960fed27512fc7',
                    'subnet-029e4362fc65caf73',
                    'subnet-01b6fb88abc34f050',
                ],
                'securityGroups': [
                    'sg-0fec3abb27c2d3dec',
                ],
                'assignPublicIp': 'DISABLED'
            }
        },
        overrides={
            'containerOverrides': [
                {
                    'name': 'task',
                    'environment': [
                        {
                            'name': name,
                            'value': value,
                        } for name, value in parameters['environment'].items()
                    ],
                    'cpu': int(parameters['cpu']),
                    'memory': int(parameters['memory']),
                    'memoryReservation': int(parameters['memory']) - 50,
                },
            ],
            'cpu': parameters['cpu'],
            'memory': parameters['memory'],
            'taskRoleArn': parameters['role']
        },
        startedBy='harbour:api',
        taskDefinition=image['definition']['arn']
    )

    event = {
        'tasks': [
            {
                'id': None,
                'az': task['availabilityZone'],
                'placement': {
                    'availability_zone': task['availabilityZone'],
                },
                'attachments': task['attachments'],
                'cluster': {
                    'arn': task['clusterArn'],
                },
                'containers': [
                    {
                        'name': container['name'],
                        'arn': container['containerArn'],
                        'task': {
                            'arn': container['taskArn'],
                        },
                        'image': container['image'],
                        'cpu': int(container['cpu']),
                        'memory_reservation': int(container['memoryReservation']),
                    }
                    for container in task['containers']
                ],
                'created_at': task['createdAt'],
                'group': task['group'],
                'launch_type': task['launchType'],
                'memory': int(task['memory']),
                'cpu': int(task['cpu']),
                'overrides': task['overrides'],
                'platform_version': task['platformVersion'],
                'started_by': task['startedBy'],
                'arn': task['taskArn'],
                'definition': task['taskDefinitionArn'],
            }
        for task in response['tasks']]
    }

    event['failures']: response['failures']

    table.update(id, event)

    response = Response(200, event)

    return response.parse()



import boto3
import json
import pymysql

from botocore.exceptions import ClientError
from harbour.database.table import Table

class Database:

    DATABASE_MAP = {
        'mysql': pymysql.connect,
    }

    def __init__(self, engine, alias, user):

        # Default parameters
        self.engine = engine
        self.alias = alias
        self.user = user

        self.credentials = None

        # Default resources
        self.client = boto3.client('secretsmanager')

    def get_connection(self):
        """Gets the credentials stored in AWS Secrets manager"""

        try:

            # Requests secrets manager for the credentials
            response = self.client.get_secret_value(
                SecretId="{}/{}/{}".format(
                    self.engine,
                    self.alias,
                    self.user
                )
            )

        except ClientError as exc:

            code = exc.response['Error']['Code']

            messages = {

                # Secrets Manager can't decrypt the protected secret text using
                # the provided KMS key.
                'DecryptionFailureException': 'Failed to decrypt the secret',
                'InternalServiceErrorException': 'Internal server error',
                'InvalidParameterException': 'Invalid parameter',
                'InvalidRequestException': 'Invalid request',
                'ResourceNotFoundException': 'Resource not found.',
                'AccessDeniedException': 'Access denied',

            }

            # Change this to logging
            print(messages.get(code))

            raise exc

        else:

            # Parses the secret
            secret = json.loads(response['SecretString'])

            # Stores the credentials
            self.credentials = {
                'host': secret['host'],
                'user': secret['username'],
                'password': secret['password'],
                'db': secret['dbname'],
                'port': int(secret['port']),
            }

            # Sets the SSDictCursor from MySQL connections
            if self.engine == 'mysql':
                self.credentials['cursorclass'] = pymysql.cursors.SSDictCursor

            # Builds the connection class, connection object and cursor object
            connect = self.DATABASE_MAP.get(self.engine)

            # Connects to the database and creates a cursor
            self.conn = connect(**self.credentials)

            self.curs = self.conn.cursor()

            return self.conn, self.curs

    def table(self, table, *args, **kwargs):
        """Returns a Table object with the Database object injected"""

        if self.credentials is None:
            self.get_connection()

        return Table(self, table)



import os
import re
import uuid

from zipfile import ZipFile, ZIP_DEFLATED

class Artifact:

    def __init__(self, *args, **kwargs):
        pass

    def package(self, path):
        """Packages the code and returns the file path"""

        # Scans the directory
        files = os.listdir(path)

        # Checks the necessary files
        self.check(files)

        # Extracts the base path from the artifact
        basepath = os.path.normpath(path)

        # Generates a unique id for the artifact
        id = uuid.uuid4().hex
        self.tmp = os.path.join('/tmp', f'{id}.zip')

        with ZipFile(self.tmp, 'w', ZIP_DEFLATED) as artifact:

            # Adds the path files
            for root, dir, files in os.walk(path):

                for file in files:

                    # Builds the filepath
                    filepath = os.path.join(root, file)

                    # Excludes filtered files
                    if self.filter(filepath) is True:
                        continue

                    # Removes the base dir
                    arcname = filepath.replace(basepath, '').replace('/','', 1)

                    # Writes the file to the artifact
                    artifact.write(filepath, arcname)

            # Adds the standard buildspec
            buildspec = os.path.join(
                os.path.dirname(__file__),
                'templates',
                'buildspec.yml'
            )

            artifact.write(buildspec, 'buildspec.yml')

        return self.tmp

    def check(self, files):
        """Checks if every necessary file is in the bundle"""

        # Checks the required build files
        REQUIRED_FILES = ['Dockerfile', 'requirements.txt']

        for file in REQUIRED_FILES:

            if file not in files:

                raise FileNotFoundError(f'The file {file} is missing')

    def filter(self, file):
        """Ignores any files matching the patterns"""

        EXCLUDE_FILES = [
            r'.*__pycache__.*', r'venv.*', r'.*cloudformation.*', r'.*\.git.*',
            r'.*\.pyc', r'.*\.python-version.*', r'.*README\.md.*', r'.*buildspec\.y[a-z]ml.*',
            r'.*Makefile.*', r'.*egg-info.*', r'.*LICENSE.*'
        ]

        for pattern in EXCLUDE_FILES:

            if re.match(pattern, file):

                return True

        return False

    def delete(self):
        """Deletes the package from the disk"""

        os.remove(self.tmp)



import hashlib
import json
import urllib
import os

import boto3

from datetime import datetime

def handler(event, context):

    codebuild = boto3.client('codebuild')

    # Messages from SQS
    for message in event['Records']:

        body = json.loads(message['body'])

        # Messages from S3
        for record in body['Records']:

            key = urllib.parse.unquote(record['s3']['object']['key'])

            # Starts to build the image
            response = codebuild.start_build(
                projectName=os.getenv('BUILDER_PROJECT_NAME'),
                environmentVariablesOverride=[
                    {
                        'name': 'IMAGE_TAG',
                        'value': key.replace('build/artifacts/', ''),
                        'type': 'PLAINTEXT'
                    },
                ],
                sourceLocationOverride=f"{record['s3']['bucket']['name']}/{key}",
                idempotencyToken=record['s3']['object']['eTag']
            )



import datetime
import os

import boto3

from decimal import Decimal

def flatten(item, lkey='', sep='.'):
    """Flattens a dict to key1.key2.key3: value notation."""
    result = {}

    for rkey, val in item.items():

        key = lkey + rkey

        if isinstance(val, dict):

            result.update(flatten(val, key + sep))

        else:

            result[key] = val

    return result


class Table:

    def __init__(self, table, pk='id', *args, **kwargs):

        # Seek order: env defined then directly defined.
        table = os.getenv(table) or table

        self.pk = pk
        self.table = boto3.resource('dynamodb').Table(table)

    def format(self, obj, *args, **kwargs):
        """Default object formatter"""

        if isinstance(obj, datetime.datetime):

            return obj.strftime('%Y-%m-%dT%H:%M:%S.%f%z')

        elif isinstance(obj, datetime.date):

            return obj.strftime('%Y-%m-%d')

        elif isinstance(obj, Decimal):

            return format(obj, '.8')

        elif isinstance(obj, bytes):

            return obj.decode(errors='ignore')

        elif isinstance(obj, datetime.timedelta):

            return str(obj)

        elif isinstance(obj, set):

            return ','.join(list(obj))

        elif isinstance(obj, list):

            return [self.format(i) for i in obj]

        elif isinstance(obj, dict):

            response = {}

            for k, v in obj.items():

                response[k] = self.format(v)

            return response

        else:

            return obj

    def get(self, pk, attributes=None, *args, **kwargs):
        """Returns the object"""

        payload = {
            'Key': {self.pk: pk},
            'ConsistentRead': True,
        }

        if attributes is not None:

            # Type checking and converting
            if isinstance(attributes, str):
                attributes = [attributes]

            elif isinstance(attributes, list):
                pass

            else:

                raise TypeError('attributes must be list or string')

            payload['ProjectionExpression'] = ', '.join(
                [f'#{a}' for a in attributes]
            )
            payload['ExpressionAttributeNames'] = {
                f'#{a}': a for a in attributes
            }

        response = self.table.get_item(**payload)

        if response['ResponseMetadata']['HTTPStatusCode'] != 200:

            # TODO: do this properly :)
            raise Exception('Deu ruim!')

        return response.get('Item', None)

    def update(self, pk, update, mode='PUT', *args, **kwargs):
        """Updates the entry on the DynamoDB"""

        if mode not in ['PUT', 'PATCH']:
            raise ValueError('Mode should be PUT or PATCH')

        payload = {
            'Key': {self.pk: pk},
            'ReturnValues': 'ALL_NEW'
        }

        # Injects the updated_at field
        update['updated_at'] = datetime.datetime.utcnow().isoformat()

        if mode == 'PUT':

            payload['UpdateExpression'] = 'SET ' + ', '.join(
                [f'#{k} = :{k}' for k in update.keys()]
            )

            payload['ExpressionAttributeNames'] = {
                f'#{k}': k for k in update.keys()
            }

            payload['ExpressionAttributeValues'] = {
                f':{k}': self.format(v) for k,v in update.items()
            }


        elif mode == 'PATCH':

            update = flatten(update)

            payload['UpdateExpression'] = 'SET ' + ', '.join(
                [f'#{k.replace(".", ".#")} = :{k.replace(".", "__")}' for k in update.keys()]
            )

            payload['ExpressionAttributeNames'] = {
                f'#{k}': k for x in update.keys() for k in x.split('.')
            }


            payload['ExpressionAttributeValues'] = {
                f':{k.replace(".", "__")}': self.format(v) for k,v in update.items()
            }


        response = self.table.update_item(**payload)

        if response['ResponseMetadata']['HTTPStatusCode'] != 200:

            # TODO: do this properly :)
            raise Exception('Deu ruim!')

        return response.get('Attributes', None)

    def delete(self, pk, *args, **kwargs):
        """Deletes a entry from DynamoDB"""

        response = self.table.delete_item(Key={self.pk: pk})

        return response

    def remove(self, pk, columns, *args, **kwargs):
        """Removes a column from DynamoDB entry"""

        payload = {
            'Key': {self.pk: pk},
            'ReturnValues': 'ALL_NEW',
        }

        if isinstance(columns, str):
            columns = [columns]

        elif isinstance(columns, list):
            pass

        else:
            raise TypeError('columns must be either string or list')

        payload['UpdateExpression'] = 'REMOVE ' + ', '.join(
            [f'#{c}' for c in columns]
        )

        payload['ExpressionAttributeNames'] = {
            f'#{c}': c for c in columns
        }

        response = self.table.update_item(**payload)

        if response['ResponseMetadata']['HTTPStatusCode'] != 200:

            # TODO: do this properly :)
            raise Exception('Deu ruim!')

        return response.get('Attributes', None)



import base64
import hashlib
import os

import boto3

from dynamodb import Table
from storage import Bucket
from api import Response
from datetime import datetime

def handler(event, context):
    """Registers a code as job"""

    method = event['httpMethod']
    resource = event['resource']

    key = (event.get('pathParameters') or {}).get('id', None)
    body = event.get('body', {})

    functions = {
        'GET': {
            '/register/{id}': detail,
            '/register': list,
        },
        'POST': {
            '/register/{id}': create,
        },
        'DELETE': {
            '/register/{id}': delete,
        },
        'PATCH': {
            '/register/{id}': update,
        },
    }

    handler = functions.get(method, {}).get(resource)

    return handler(key=key, body=body)


def detail(key, *args, **kwargs):
    """Describes a entry"""

    table = Table('REGISTRY_TABLE_NAME')

    response = Response(200, table.get(key))

    return response.parse()

def list(*args, **kwargs):
    """Lists all entries"""
    pass

def delete(key, *args, **kwargs):
    """Deletes a entry"""
    pass

def update(key, body, *args, **kwargs):
    """Updates (patches) a entry"""
    pass

def create(key, body, *args, **kwargs):
    """Creates a entry"""

    # Creates the S3 client
    client = boto3.client('s3')
    bucket = Bucket('REGISTRY_BUCKET_NAME')
    table = Table('REGISTRY_TABLE_NAME')

    # Generates the file md5
    hash = hashlib.md5(key.encode()).digest().hex()

    # Gets the content of the payload
    filename = os.path.join('/tmp', f'{hash}.zip')
    filekey = f'build/artifacts/{hash}'

    with open(filename, 'wb+') as file:
        file.write(base64.b64decode(body))

    # Writes the content to S3
    bucket.upload(filename, filekey, delete=True)

    # Returns the object id
    object = {
        'name': key,
        'status': 'UPLOADED',
        'created_at': datetime.utcnow().isoformat(),
        'location': f's3://{bucket.name}/{filekey}',
    }

    table.update(hash, object)

    response = Response(200, {'id': hash, 'object': object})

    return response.parse()



class Table:

    def __init__(self, db, name, *args, **kwargs):

        self.schema = db.credentials['db']
        self.name = name

        # Default state
        self.sql = None
        self.columns = '*'
        self.filters = {}
        self._limit = None
        self.order = None
        self.primary_key = None

        # Database connections
        self.db = db
        self.conn, self.curs = db.conn, db.curs
        self.host = db.credentials['host']

    def select(self, columns, *args, **kwargs):
        """Defines the columns selected"""

        # Type checking
        if columns == '*':

            self.columns = ['*']

        elif isinstance(columns, str):

            self.columns = [columns]

        elif not isinstance(columns, list):

            raise TypeError('columns must be str or list')

        return self

    def order_by(self, order, *args, **kwargs):
        """Appends a ORDER BY clausule to the query"""

        if not isinstance(order, dict):
            raise TypeError('order must be a dict {"column": "ASC|DESC"}')

        for value in order.values():

            if value not in ['ASC', 'DESC']:
                raise AttributeError('order directoin must be ASC or DESC')

        self.order = order

        return self

    def where(self, *filters):
        """Appends a where clausule to the query.
        Premisses:
            lists are converted to IN() statements.
            strings and ints are converted to = statements.
        """

        for f in filters:
            self.filters.update(f)

        return self

    def limit(self, limit, *args, **kwargs):
        """Appends a LIMIT clausule to the query"""

        if not isinstance(limit, int):
            raise TypeError('limit must be a int')

        if limit != 0:

            self._limit = limit

        return self

    def build(self, *args, **kwargs):
        """Builds the SQL"""

        self.sql = "/*Application: Tethys*/\n"
        self.sql += "SELECT {}".format(', '.join(self.columns))
        self.sql += ", CAST(UNIX_TIMESTAMP() * POW(10, 6) AS DECIMAL(25,9)) AS _sequence\n"
        self.sql += "FROM {}.{}\n".format(self.schema, self.name)

        if self.filters != {}:

            self.sql += "WHERE\n\t1=1\n"

            for c, f in self.filters.items():

                if isinstance(f, list):

                    self.sql += "\tAND {} IN({})\n".format(
                        c, ','.join([repr(v) for v in f])
                    )

                if isinstance(f, (str, int,)):

                    self.sql += "\tAND {} = {}\n".format(
                        c, repr(f)
                    )

        if self.order is not None:

            self.sql += "ORDER BY {}\n".format(
                ','.join(['{} {}'.format(k,v) for k, v in self.order.items()])
            )

        if self._limit is not None:
            self.sql += "LIMIT {}".format(self._limit)

        return self

    def execute(self):
        """Executes a query to check the table metadata and then executes the
        built sql"""

        self.build()

        self.curs.execute(self.sql)

        return self

    @property
    def fields(self):
        """Returns the table schema"""

        MYSQL_TYPE_MAP = {
            0: 'float64', 1: 'int2', 2: 'int2', 3: 'int4', 4: 'float64',
            5: 'float64', 6: 'null', 7: 'timestamp', 8: 'int64', 9: 'int64',
            10: 'date', 11: 'time', 12: 'datetime', 13: 'year', 14: 'date',
            15: 'string', 16: 'int2', 17: 'datetime', 18: 'datetime',
            19: 'time', 245: 'string', 246: 'float64', 247: 'string',
            248: 'string', 249: 'string', 250: 'string', 251: 'string',
            252: 'string', 253: 'string', 254: 'string', 255: 'string',
        }

        fields = {
            field[0]: {
                'type': MYSQL_TYPE_MAP.get(field[1])
            } for field in self.curs.description if field[0] != '_sequence'
        }

        return fields

    @property
    def pk(self):
        """Returns the table primary key"""

        # This a cache. Do not remove!
        if self.primary_key is None:

            sql = f"""
            SELECT
                u.COLUMN_NAME
            FROM information_schema.TABLE_CONSTRAINTS c
            INNER JOIN information_schema.KEY_COLUMN_USAGE u
                ON c.TABLE_SCHEMA = u.TABLE_SCHEMA
                AND c.TABLE_NAME = u.TABLE_NAME
                AND c.CONSTRAINT_NAME = u.CONSTRAINT_NAME
            WHERE
                c.TABLE_SCHEMA = '{self.schema}'
                AND c.TABLE_NAME = '{self.name}'
                AND c.CONSTRAINT_TYPE = 'PRIMARY KEY'
            ORDER BY u.ORDINAL_POSITION ASC;
            """

            with self.conn as cursor:

                cursor.execute(sql)

                self.primary_key = [
                    row['COLUMN_NAME'] for row in cursor.fetchall()
                ]

        return self.primary_key

    def fetchall(self):
        """Fetches all the records."""

        for row in self.curs.fetchall():
            yield row

    def fetchmany(self, n=1000):
        """Fetches as many records as specified. Defaul: 1000"""

        return self.curs.fetchmany(n)

    def __iter__(self):
        """Iterates over the table records."""

        return self.fetchall()



import setuptools

setuptools.setup(
    name="harbour",
    version="0.0.12",
    author="ebanx-dataops",
    author_email="dataops@ebanx.com",
    description="Running docker on AWS made easy",
    url="https://github.com/ebanx/harbour",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    entry_points={
        "console_scripts": [
            "harbour = harbour.cli:cli"
        ]
    },
    install_requires=[
        'click==7.0',
        'requests==2.22.0',
        'boto3==1.11.9',
        'structlog==20.1.0',
        'PyMySQL==0.9.3',
    ],
    extras_require={
        'test': [],
        'build': [
            'twine',
        ],
    }
)



import click

from harbour.client import Client
from harbour.client.artifact import Artifact

@click.group()
def cli():
    pass

@cli.command()
def configure():
    """Configures the credentials file"""

    client = Client()

    domain = click.prompt('Habour domain', type=str)
    key = click.prompt('Habour key', type=str)

    client.configure(domain, key)

@cli.command()
@click.option('--path', required=True, help='Root directory of the project')
@click.option('--name', required=True, help='Image name:tag. If tag is ommited, latest is considered.')
def register(**kwargs):
    """Uploads a code to the repository"""

    client = Client()

    result = client.register(**kwargs)


if __name__ == '__main__':
    cli()


@cli.command()
@click.option('--cpu', required=False, help='Execution CPU override')
@click.option('--memory', required=False, help='Execution memory override')
@click.option('--role', required=False, help='Execution role override')
@click.option('--environment', '-e', multiple=True, help='Overrides the environment variable. Format: NAME=value')
@click.argument('image')
def execute(*args, **kwargs):
    """Executes a image"""

    client = Client()

    # Parses the environment variables
    environment = {}

    for env in kwargs.get('environment', []):

        key, value = env.split('=')

        environment[key] = value

    kwargs['environment'] = environment

    result = client.execute(**kwargs)


@cli.command()
@click.argument('path')
def package(path):
    """Packages a distribution"""

    artifact = Artifact()

    filepath = artifact.package(path)

    click.echo(filepath)



import traceback

import structlog

structlog.configure(
    processors=[
        structlog.processors.TimeStamper(fmt='iso'),
        structlog.stdlib.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.JSONRenderer(),
    ],
)

class Logger:

    def __init__(self):

        self.log = structlog.get_logger()

    def parse(self, message):
        """
        Parses the message to a dict. This allows for ES to index it properly
        """

        if isinstance(message, str):
            message = {
                'text': message,
            }

        return message

    def critical(self, message):
        self.log.critical(self.parse(message))

    def debug(self, message):
        self.log.debug(self.parse(message))

    def error(self, message):
        self.log.error(self.parse(message))

    def fatal(self, message):
        self.log.fatal(self.parse(message))

    def info(self, message):
        self.log.info(self.parse(message))

    def msg(self, message):
        self.log.msg(self.parse(message))

    def warning(self, message):
        self.log.warning(self.parse(message))

    def metric(self, name, value):
        self.log.msg({'metric': name, 'value': value})

    def bind(self, **parameters):
        self.log = self.log.bind(**parameters)

    def unbind(self, *keys):
        self.log = self.log.unbind(keys)

    def exception(self, exc):

        tb = traceback.format_exc().replace('\n', ';;')

        self.log.fatal({'exception': exc, 'traceback': tb})
